using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Configuration;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using System.Text.RegularExpressions;
using System.Xml;


namespace SqlFixture
{
	public class TestInformation
	{
		protected const string DefaultCase = "__DEFAULT__";
		string _filename;
		string _key;
		Dictionary<string, Dictionary<string, Dictionary<string, object>>> _settings;
		Dictionary<string, Dictionary<string, Dictionary<string, IExpected>>> _expected;

		#region Static Methods & Properties
		public static NameValueCollection Configuration
		{
			get
			{
				NameValueCollection col = ConfigurationManager.GetSection("SqlFixture/TestInformation") as NameValueCollection;
				if (col == null) { throw new ConfigurationErrorsException("SqlFixture / TestInformation configuration section not found."); }
				return col;
			}
		}
		#endregion

		#region Constructors & Initialization
		public TestInformation(string key)
		{
			_key = key;
			LoadFile(Configuration[key]);
		}

		protected void LoadFile(string filename)
		{
			// Open file based on current application path
			string filePath = System.IO.Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().GetName().CodeBase);
			filePath = Regex.Replace(filePath, @"^[a-z0-9]+:\\", "", RegexOptions.IgnoreCase);
			_filename = Path.Combine(filePath, filename);
			if (!File.Exists(_filename)) { throw new ConfigurationErrorsException(String.Format("Cannot find expected values file {0}", _filename)); }
			FileStream fs = new FileStream(_filename, FileMode.Open, FileAccess.Read);

			// Read as XML and load into class
			LoadXml(fs);
			fs.Close();
		}
		#endregion

		#region XML Parsing
		protected const string DuplicateFixtureError = "Duplicate Fixture name {0} in file {1}";
		protected const string UnnamedFixtureError = "Unnamed Fixture in file {0}";
		protected const string DuplicateTestError = "Duplicate Test name {0} in Fixture {1}, file {2}";
		protected const string UnnamedTestError = "Unnamed Test in Fixture {0}, file {1}";
		protected const string DuplicateSettingError = "Duplicate Setting name {0} in Fixture {1}, Test {2}, file {3}";
		protected const string UnnamedSettingError = "Unnamed Setting in Fixture {0}, Test {1}, file {2}";
		protected const string DuplicateExpectedError = "Duplicate Expected name {0} in Fixture {1}, Test {2}, file {3}";
		protected const string InvalidTypeError = "Type {0} not recognized in file {1}";
		protected const string TooManyValuesExpectedError = "{0} expected values given for Expected name {1} in fixture {2}, Test {3}, file {4}";

		protected void LoadXml(Stream stream)
		{
			XmlDocument doc = new XmlDocument();
			doc.Load(stream);
			_settings = new Dictionary<string, Dictionary<string, Dictionary<string, object>>>();
			_expected = new Dictionary<string, Dictionary<string, Dictionary<string, IExpected>>>();

			// Loop through fixtures
			foreach (XmlElement fixtureElement in doc.DocumentElement.GetElementsByTagName("Fixture"))
			{
				// If fixture name is not present or is duplicated, throw an error
				string fixtureName = fixtureElement.HasAttribute("name") ? fixtureElement.GetAttribute("name") : "";
				if (String.IsNullOrEmpty(fixtureName)) { throw new ConfigurationErrorsException(String.Format(UnnamedFixtureError, _filename)); }
				if (_settings.ContainsKey(fixtureName)) { throw new ConfigurationErrorsException(String.Format(DuplicateFixtureError, fixtureName, _filename)); }
				
				// Create key in settings and expected values for fixture
				_settings[fixtureName] = new Dictionary<string, Dictionary<string, object>>();
				_expected[fixtureName] = new Dictionary<string, Dictionary<string, IExpected>>();

				// Loop through tests
				foreach (XmlElement testElement in fixtureElement.GetElementsByTagName("Test"))
				{
					// If test name is not present or is duplicated, throw an error
					string testName = testElement.HasAttribute("name") ? testElement.GetAttribute("name") : "";
					if (String.IsNullOrEmpty(testName)) { throw new ConfigurationErrorsException(String.Format(UnnamedFixtureError, fixtureName, _filename));  }
					if (_settings[fixtureName].ContainsKey(testName)) { throw new ConfigurationErrorsException(String.Format(DuplicateFixtureError, testName, fixtureName, _filename)); }

					// Create key in settings and expected values for test
					_settings[fixtureName][testName] = new Dictionary<string, object>();
					_expected[fixtureName][testName] = new Dictionary<string, IExpected>();

					// Loop through settings
					foreach (XmlElement settingElement in testElement.GetElementsByTagName("Setting"))
					{
						// If setting name is not present or is duplicated, throw an error
						string settingName = settingElement.HasAttribute("name") ? settingElement.GetAttribute("name") : "";
						if (String.IsNullOrEmpty(settingName)) { throw new ConfigurationErrorsException(String.Format(UnnamedSettingError, fixtureName, testName, _filename)); }
						if (_settings[fixtureName][testName].ContainsKey(settingName)) { throw new ConfigurationErrorsException(String.Format(DuplicateSettingError, settingName, fixtureName, testName, _filename)); }
						
						string typeName = settingElement.HasAttribute("type") ? settingElement.GetAttribute("type") : "string";
						
						// Parse value and store
						string settingValue = (settingElement.ChildNodes.Count > 0) ? settingElement.ChildNodes[0].Value : null;
						_settings[fixtureName][testName][settingName] = GetValue(typeName, settingValue);
					}

					// Loop through expected values
					foreach (XmlElement expectedElement in testElement.GetElementsByTagName("Expected"))
					{
						// If case name is duplicated, throw an error
						string caseName = expectedElement.HasAttribute("name") ? expectedElement.GetAttribute("name") : DefaultCase;
						string typeName = expectedElement.HasAttribute("type") ? expectedElement.GetAttribute("type") : null;
						if (_expected[fixtureName][testName].ContainsKey(caseName)) { throw new ConfigurationErrorsException(String.Format(DuplicateExpectedError, caseName, fixtureName, testName, _filename)); }
						
						// Handle an expected object
						XmlNodeList expectedObjectNodes = expectedElement.GetElementsByTagName("Object");
						List<IExpected> expected = new List<IExpected>();
						foreach (XmlElement expectedObjectNode in expectedObjectNodes)
						{
							expected.Add(new ExpectedObject(expectedObjectNode));
						}
						
						// Handle string values
						foreach (XmlNode node in expectedElement.ChildNodes)
						{
							if (node.NodeType == XmlNodeType.Text && !String.IsNullOrEmpty(node.InnerText.Trim()))
							{
								if (typeName != null) { expected.Add(new ExpectedStringValue(node.InnerText.Trim(), typeName)); }
								else { expected.Add(new ExpectedStringValue(node.InnerText.Trim())); }
							}
						}

						// Parse value and store
						if (expected.Count > 1) { throw new ConfigurationErrorsException(String.Format(TooManyValuesExpectedError, expected.Count, caseName, fixtureName, testName, _filename)); }
						_expected[fixtureName][testName][caseName] = (expected.Count > 0) ? expected[0] : new ExpectedNull();
					}
				}
			}
		}

		protected object GetValue(string typeName, string value)
		{
			switch (typeName.ToLower())
			{
				case "byte":
					return Byte.Parse(value);

				case "sbyte":
					return SByte.Parse(value);

				case "int":
					return int.Parse(value);

				case "int16":
					return Int16.Parse(value);

				case "int32":
					return Int32.Parse(value);

				case "int64":
					return Int64.Parse(value);

				case "uint16":
					return UInt16.Parse(value);

				case "uint32":
					return UInt32.Parse(value);

				case "uint64":
					return UInt64.Parse(value);

				case "single":
					return Single.Parse(value);

				case "double":
					return Double.Parse(value);

				case "boolean":
					return Boolean.Parse(value);

				case "char":
					return Char.Parse(value);

				case "decimal":
					return Decimal.Parse(value);

				case "string":
					return value;
				
				case "datetime":
					return DateTime.Parse(value);

				default:
					return value.ToString();
			}
		}
		#endregion

		#region Retrieve Values
		private CallerInfo GetCaller()
		{
			StackTrace stackTrace = new StackTrace();
			StackFrame stackFrame; MethodBase methodBase;
			for (int i = 0; i < stackTrace.FrameCount; i++)
			{
				stackFrame = stackTrace.GetFrame(i);
				methodBase = stackFrame.GetMethod();
				if (methodBase.DeclaringType != this.GetType())
				{
					foreach (Attribute attr in methodBase.GetCustomAttributes(true))
					{
						string typeName = attr.GetType().Name;
						if (typeName == "TestAttribute")
						{
							return new CallerInfo(methodBase);
						}
					}
				}
			}
			return new CallerInfo();
		}

		public T GetSettingValue<T>(string fixtureName, string testName, string settingName)
		{
			if (!_settings.ContainsKey(fixtureName)) { throw new KeyNotFoundException(String.Format("Fixture {0} not found in {1} test settings.", fixtureName, _key)); }
			if (!_settings[fixtureName].ContainsKey(testName)) { throw new KeyNotFoundException(String.Format("Test {0} not found in fixture {1} in {2} test settings.", testName, fixtureName, _key)); }
			if (!_settings[fixtureName][testName].ContainsKey(settingName)) { throw new KeyNotFoundException(String.Format("Setting {0} not found in fixture {1}, test {2} in {3} test settings", settingName, fixtureName, testName, _key)); }
			object returnValue = _settings[fixtureName][testName][settingName];
			return (returnValue is T) ? (T)returnValue : default(T);
		}

		public T GetSettingValue<T>(string settingName)
		{
			CallerInfo callerInfo = GetCaller();
			return GetSettingValue<T>(callerInfo.Class, callerInfo.Method, settingName);
		}

		public IExpected GetExpectedValue(string fixtureName, string testName, string caseName)
		{
			if (!_expected.ContainsKey(fixtureName)) { throw new KeyNotFoundException(String.Format("Fixture {0} not found in {1} test settings.", fixtureName, _key)); }
			if (!_expected[fixtureName].ContainsKey(testName)) { throw new KeyNotFoundException(String.Format("Test {0} not found in fixture {1} in {2} test settings.", testName, fixtureName, _key)); }
			if (!_expected[fixtureName][testName].ContainsKey(caseName)) { throw new KeyNotFoundException(String.Format("Case {0} not found in fixture {1}, test {2} in {3} test settings.", caseName, fixtureName, testName, _key)); }
			return _expected[fixtureName][testName][caseName];
		}

		public IExpected GetExpectedValue(string fixtureName, string testName)
		{
			return GetExpectedValue(fixtureName, testName, DefaultCase);
		}

		public IExpected GetExpectedValue(string caseName)
		{
			CallerInfo callerInfo = GetCaller();
			return GetExpectedValue(callerInfo.Class, callerInfo.Method, caseName);
		}

		public IExpected GetExpectedValue()
		{
			return GetExpectedValue(DefaultCase);
		}
		#endregion
	}
}
